"""Mirador API client

.. moduleauthor:: Nick Jacob <nick@mirador.im>

"""
from __future__ import absolute_import

import base64
import requests
from .errors import http_exceptions, MiradorException
from .result import MiradorResult


def import_async_requests():
    try:
        from grequests import async
        return async
    except ImportError:
        try:
            from requests import async
            return async
        except ImportError:
            pass

        raise MiradorException(
            "grequests async library required for asynchronous requests"
        )


class MiradorClient(object):
    """Interface for the Mirador API

    Simple interface for retrieving results from the
    Mirador API. Returns [MiradorResult] on all method
    """

    API_BASE = "http://api.mirador.im"
    CLASSIFY_ENDPOINT = "/v1/classify"
    TIMEOUT = 10
    HEADERS = {
        'User-Agent': 'MiradorClient/1.0 Python'
    }

    def __init__(self, api_key, timeout=10):
        """Instaniate a MiradorClient
        Args:
            api_key: the mirador api key (string)
            timeout: timeout for API access (default: 10 seconds)

        Returns:
            A MiradorClient instance
        """

        self.TIMEOUT = timeout or self.TIMEOUT

        self._api_key = api_key
        self._url = "{0}{1}".format(
            self.API_BASE, self.CLASSIFY_ENDPOINT
        )

    def _params(self, data, method):
        """prepare paramters for the request"""

        key = 'data' if method == 'post' else 'params'
        data['api_key'] = self._api_key

        return {'headers': self.HEADERS, key: data}

    def _prepare_request(self, **data):
        """prepare the options & parameters of request"""
        if not data or ('image' not in data and 'url' not in data):
            raise http_exceptions[400]("url(s) or image(s) required")

        method = 'get' if not 'image' in data else 'post'
        params = self._params(data, method)

        return method, params

    def _request(self, **data):
        """make the request via requests module"""

        method, params = self._prepare_request(**data)
        r = getattr(requests, method)(self._url, **params)

        if r.status_code != 200:
            raise http_exceptions[r.status_code](r.text)
        return r.json()

    def _async_request(self, files_or_urls, on_done, **data):
        """make an asynchronous request to the API"""
        async = import_async_requests()

        # very basic callback that just
        # parses the response and hands it off
        # to the handler
        def handle_async_response(response):
            if response.status_code == 200:
                on_done(
                    MiradorResult.parse_results_safe(
                        files_or_urls,
                        response.json().get('results', None)
                    )
                )

        method, params = self._prepare_request(**data)

        # set up our response hook
        params['hooks'] = {'response': handle_async_response}
        r = getattr(async, method)(self._url, **params)

        # execute the request
        async.map([r])

    def _read_image(self, image):
        """read an image if it's a file and return the name"""

        if isinstance(image, basestring):
            with open(image, 'rb') as imf:
                return image, imf.read()
        else:
            return image.name, image

    def _prepare_image(self, im_f):
        """convert an image name or file into base64 encoded string"""

        name, data = self._read_image(im_f)

        if not data:
            raise http_exceptions[400]("no image data for: {}".format(name))
        return base64.b64encode(data)

    def classify_urls(self, *urls):
        """classifies url(s) via the API

        Args:
            *urls: urls to classify. Must be publically accessible
        Returns:
            a list of MiradorResult objects, each with three properties:
                name: a string representing the source url
                safe: boolean - is the image safe (unflagged)
                value: float representing confidence in flag result
        """
        if not urls:
            return []

        if isinstance(urls[0], (list, tuple)):
            urls = urls[0]

        res = self._request(url=urls)

        return MiradorResult.parse_results(
            urls, res.get('results', None)
        )

    def classify_files(self, *files):
        """classifies file(s) via the API
        Args:
            *files: filenames or file objects
        Returns:
            a list of MiradorResult objects, each with three properties:
                name: a string representing the source filename
                safe: boolean - is the image safe (unflagged)
                value: float representing confidence in flag result
        """
        if not files:
            return []

        if isinstance(files[0], (list, tuple)):
            files = files[0]

        res = self._request(
            image=map(self._prepare_image, files)
        )

        return MiradorResult.parse_results(
            files, res.get('results', None)
        )

    def async_classify_files(self, files, on_done):
        """asynchronously classify files, using `on_done` callback"""
        self._async_request(image=map(self._prepare_image, files))



# test the mirador python package
import unittest
from test_client import suite as client_suite


def suite():
    return unittest.TestSuite([client_suite])


if __name__ == "__main__":
    unittest.main()



# Mirador python client
# @author nickjacob (nick@mirador.im)
from client import MiradorClient
from errors import MiradorException
from os import getenv as _genv

import client
import ext


def classify_files(files, api_key=None):
    "procedural: classify files given the `api_key`"
    api_key = api_key or _genv('MIRADOR_API_KEY')

    if not api_key:
        raise MiradorException('no api key provided')
    return MiradorClient(api_key).classify_files(*files)


def classify_urls(urls, api_key=None):
    "procedural: classify urls given the `api_key`"
    api_key = api_key or _genv('MIRADOR_API_KEY')

    if not api_key:
        raise MiradorException('no api key provided')
    return MiradorClient(api_key).classify_urls(*urls)



# exceptions for access to the Mirador API
# subclassing appropriate exceptions for function
from httplib import HTTPException
from collections import defaultdict


class MiradorException(HTTPException):
    code = 500

    def __init__(self, *args, **kwargs):
        super(MiradorException, self).__init__(
            "[{code}]: {message}".format(
                message=" ".join(map(str, args)), code=self.code)
        )


class InvalidRequestException(MiradorException):
    code = 400

    def __init__(self, *args, **kwargs):
        super(InvalidRequestException, self, *args, **kwargs).__init__(
            "must include valid URL or file"
        )


class UnauthorizedException(MiradorException):
    code = 403

    def __init__(self, *args, **kwargs):
        super(UnauthorizedException, self).__init__(
            "API key is invalid. Contact support@mirador.im for valid key"
        )


def default_exception():
    return MiradorException


http_exceptions = defaultdict(default_exception)
http_exceptions[400] = InvalidRequestException
http_exceptions[403] = UnauthorizedException
http_exceptions[500] = MiradorException



"""Mirador Result class

.. moduleauthor Nick Jacob <nick@mirador.im>

"""
from errors import http_exceptions, MiradorException


class MiradorResult(object):
    """A result from the Mirador API.
    Contains fields indicating classification result:
        safe - boolean indicating flagging status
        value - a float 0.0-1.0; confidence of result
    """

    FMT_STR = "<MiradorResult: {name}; safe: {safe}; value: {value}/>"

    def __init__(self, name, raw={}):

        if 'result' not in raw:
            raise http_exceptions[500]("bad result: {}".format(raw))

        res = raw['result']

        if not 'safe' in res and 'value' in res:
            raise http_exceptions[500]("bad result: {}".format(raw))

        self.name = name
        self.safe = res['safe']
        self.value = res['value']

    @staticmethod
    def parse_results(reqs, results):
        """ parse JSON output of API into MiradorResult objects """

        if not results:
            raise http_exceptions[500]("no result available: {}".format(reqs))

        return [
            MiradorResult(n, r)
            for n, r in zip(reqs, results)
        ]

    @staticmethod
    def parse_results_safe(reqs, results):
        "parse results and catch errors in bad results"
        if not results:
            return None

        try:
            return [
                MiradorResult(n, r)
                for n, r in zip(reqs, results)
            ]
        except MiradorException:
            return None

    def __repr__(self):
        return self.FMT_STR.format(**self.__dict__)

    def __str__(self):
        return self.FMT_STR.format(**self.__dict__)



import unittest
from mirador.client import MiradorClient
from os import getenv as _genv
from os import path as _p

MIRADOR_API_KEY = _genv('MIRADOR_API_KEY')
__all__ = (
    'TestMiradorClient'
)


class TestMiradorClient(unittest.TestCase):

    def setUp(self):
        self.client = MiradorClient(
            MIRADOR_API_KEY
        )

    def test_classify_urls(self):
        "check that classification works both on server & client"

        results = self.client.classify_urls(
            'http://mirador.im/test/nsfw.jpg',
            'http://mirador.im/test/sfw.jpg'
        )

        self.assertEqual(len(results), 2)
        nsfw, sfw = results

        # check that the filenames were matched correctly
        self.assertEqual(nsfw.name, 'http://mirador.im/test/nsfw.jpg')
        self.assertEqual(sfw.name, 'http://mirador.im/test/sfw.jpg')

        # check that the classification was correct
        self.assertEqual(nsfw.safe, False)
        self.assertGreaterEqual(nsfw.value, 0.50)

        self.assertEqual(sfw.safe, True)
        self.assertLessEqual(sfw.value, 0.50)

    def test_classify_files(self):
        dirname = _p.dirname(__file__)

        nsfw_filename = _p.join(dirname, 'images/nsfw.jpg')
        sfw_filename = _p.join(dirname, 'images/sfw.jpg')

        test_file = open(sfw_filename, 'rb')

        # get the results from API
        results = self.client.classify_files(
            nsfw_filename, test_file)

        self.assertEqual(len(results), 2)
        nsfw, sfw = results

        self.assertEqual(nsfw.name, nsfw_filename)
        self.assertEqual(sfw.name, sfw_filename)

        # check that the results are correct
        self.assertEqual(nsfw.safe, False)
        self.assertGreaterEqual(nsfw.value, 0.50)

        self.assertEqual(sfw.safe, True)
        self.assertLessEqual(sfw.value, 0.50)


def suite():
    suite = unittest.TestLoader().loadTestsFromTestCase(TestMiradorClient)
    return suite



from distutils.core import setup

setup(
    name='Mirador',
    version='0.1.4',
    author='Nick Jacob - Mirador',
    author_email='nick@mirador.im',
    packages=['mirador', 'mirador.test', 'mirador.ext', 'mirador.errors'],
    scripts=['bin/mirador-client'],
    url='http://github.com/mirador-cv/mirador-py',
    license='LICENSE.txt',
    description='client for the Mirador image moderation API. (mirador.im)',
    install_requires=[
        "requests >= 2.3.0"
    ],
)



from __future__ import absolute_import
from ..errors import MiradorException
from ..client import MiradorClient

try:
    import django.db
    import django.db.models
    from django.db.models.signals import pre_save
    from django.dispatch import receiver
except ImportError:
    print "error: failed to import django"
    raise

# only want to expose moderate_field
# when someone goes after the whole package
__all__ = (
    'moderate_field'
)


def get_model_image(instance, image_field_name):
    "retrieve the image object from the model safely"

    if not instance:
        return None

    if image_field_name not in instance:
        raise MiradorException(
            "field '%s' not in instance" % image_field_name)

    image = instance[image_field_name]

    if not image or not getattr(image, 'read', None):
        return None

    return image


def classify_file(client, file, output_type=float):
    "use the mirador client to classify the file, parse result"

    result = client.classify_files(file)
    if not result:
        return None

    return result.value if output_type == float else result.safe


def moderate_field(model=None, image_field_name=None,
                   moderation_field_name=None,
                   moderation_field_type=float,
                   on_moderation=None, api_key=''):
    """
    Attach to a model's `pre_save` hook, updating the
    `moderation_field_name` with the value of the Mirador
    API result. This field is a float by default, but can also
    be a boolean indicating flagging (based on our threshold).
    For asynchronous results, provide a callback for `on_moderation`
    instead of a `moderation_field.` This will be called with the following
    parameters::
        def on_moderation_cb(model_instance, is_safe, value):
            "is_safe - boolean, value - float"
    """

    print "attaching hook"

    @receiver(pre_save, sender=model)
    def on_presave_signal(sender, **kwargs):

        if not 'instance' in kwargs:
            return None


        instance = kwargs['instance']
        image = get_model_image(
            instance, image_field_name)

        print "received instance: {}".format(instance)

        if not instance or not image:
            return None

        mirador_client = MiradorClient(api_key)

        # if the user has passed in a callback, use this by default
        # and do not update the original model. This will use an
        # asynchronous request in `requests`.
        if on_moderation is not None:

            def moderation_cb_wrapper(results):
                if not results or len(results) != 1:
                    return on_moderation(image, None, None)

                on_moderation(instance, results[0].safe, results[0].value)

            mirador_client.async_classify_files([image], moderation_cb_wrapper)
            return None

        print "classifying image: {}".format(image)

        # if this fails, the output will be `None`
        col_data = classify_file(
            mirador_client, image, moderation_field_type)

        # put the data onto the instance; will be saved in the db
        instance[moderation_field_name] = col_data
        return col_data
